#include "confdir.h"

#define cd_errchan stderr
#define cd_error(msg) fprintf(cd_errchan, "%s\n", msg)

void cd_getenvinfo(confdirinfo * cdi)
{
	/* wywoywane dla aplikacji majcej kontakt ze rodowiskiem uytkownika */
	cdi->env.config_dir = getenv("CONFIG_DIR");
	cdi->env.data_dir = getenv("DATA_DIR");
	cdi->env.home = getenv("HOME");
	if (cdi->env.home == NULL)
		cd_error("HOME environment string not set!");
}

int cd_splitline(char * line, char *** tokens, int len)
{
	/* dzieli lini tekstu na wyrazy, traktujc "aa aa " jako jeden wyraz */
	int n = 0, i = 0, j;
	char ** tmp;
	char delim;
	*tokens = (char **) realloc(*tokens, 0);
	while (i < len) {
		if (line[i] == ' ')
			i++;
		else {
			switch (line[i]) {
				case '\"': delim = '\"'; break;
				case '\'': delim = '\''; break;
				default: delim = ' '; i--;
			}
			j = i + 1;
			while (j < len && line[j] != delim)
				j++;
			if (line[j] != delim)
				break;
			tmp = (char **) realloc((void *) *tokens, sizeof(char *) * (n + 1));
			if (!tmp)
				return n;
			*tokens = tmp;
			(*tokens)[n] = (char *) malloc(sizeof(char) * (j - i));
			if (!(*tokens)[n]) {
				tmp = (char **) realloc((void *) *tokens, sizeof(char *) * n);
				if (!tmp)
					return -1;
				return n;
			}
			memcpy((void *) (*tokens)[n], line + i + 1, (j - i) * sizeof(char));
			((* tokens)[n])[j - i - 1] = 0;
			n++;
			i = j + 1;
		}
	}
	return n;
}

int cd_getfileinfo(confdirinfo * cdi)
{
	/* wywoywane dla aplikacji nie majcej kontaktu ze rodowiskiem uytkownika */
	size_t l;
	FILE * f;
	char etcrc_path[MAXPATHLEN + 1];
	char * line = NULL;
	char ** tokens;
	size_t len = 0;
	ssize_t read;
	l = (strlen(cdi->home_dir) + 1) * sizeof(char);
	cdi->env.home = (char *) malloc(l);
	if (cdi->env.home == NULL)
		cd_error("Memory allocation error!");
	else
		memcpy((void *) cdi->env.home, (const void *) cdi->home_dir, l);
	memcpy((void *) etcrc_path, (const void *) cdi->home_dir, MAXPATHLEN * sizeof(char));
	l = cd_purifyname(etcrc_path);
	cd_appendsubname(etcrc_path, CD_ETCRC, l);
	if ((f = fopen(etcrc_path, "r")) == NULL) {
		fprintf(cd_errchan, "Using global configuration paths file %s\n", CD_GLOBAL_ETCRC);
		if ((f = fopen(CD_GLOBAL_ETCRC, "r")) == NULL) {
			fprintf(cd_errchan, "Could not open %s: %s\n", CD_GLOBAL_ETCRC, strerror(errno));
			return -1;
		}
	}
	cdi->env.config_dir = NULL;
	cdi->env.data_dir = NULL;
	while ((read = getline(&line, &len, f)) != -1) {
		if (line != NULL && (line[0] == '#' || !read))
			continue;
		if (line[read - 1] == '\n') {
			read--;
			line[read] = '\0';
		}
		if (cd_splitline(line, &tokens, read) < 2)
			continue;
		if (!strcmp(tokens[0], "CONFIG_DIR")) {
			l = strlen(tokens[1]) + 1;
			cdi->env.config_dir = (char *) malloc(sizeof(char) * l);
			if (!cdi->env.config_dir)
				cd_error("Memory allocation error!");
			else
				strncpy(cdi->env.config_dir, tokens[1], l);
		}
		if (!strcmp(tokens[0], "DATA_DIR")) {
			l = strlen(tokens[1]) + 1;
			cdi->env.data_dir = (char *) malloc(sizeof(char) * l);
			if (!cdi->env.data_dir)
				cd_error("Memory allocation error!");
			else
				strncpy(cdi->env.data_dir, tokens[1], l);
		}
	}
	if (line)
		free(line);
	fclose(f);
	return 0;
}

void cd_freestrings(confdirinfo * cdi)
{
	/* wywoywane, jeli wczeniej byo wywoane cd_getfileinfo */
	if (cdi->env.config_dir != NULL)
		free(cdi->env.config_dir);
	if (cdi->env.data_dir != NULL)
		free(cdi->env.data_dir);
	free(cdi->env.home);
}

int cd_purifyname(char dirname[MAXPATHLEN + 1])
{
	int i, j, k, l;
	i = 0;
	while (i < MAXPATHLEN && dirname[i] != '\0') {
		if (dirname[i] == '/') {
			l = i;
			j = i+1;
			while (j < MAXPATHLEN && dirname[j] == '/')
				j++;
			if (j > i + 1) {
				i++;
				for (k = 0; j + k < MAXPATHLEN; k++)
					dirname[i + k] = dirname[j + k];
			} else
				i++;
		} else
			i++;
	}
	if (dirname[l + 1] == '\0') {
		dirname[l] = '\0';
		return l;
	} else
		return i;
}

void cd_appendsubname(char topdir[MAXPATHLEN + 1], const char * subname, size_t topdirlen)
{
	topdir[topdirlen] = '/';
	topdir[topdirlen + 1] = '\0';
	strncat(topdir, subname, MAXPATHLEN - topdirlen - 1);
	cd_purifyname(topdir);
}

int cd_rename(const char * oldpath, const char * newpath)
{
	FILE * in, * out;
	int c;
	if ((in = fopen(oldpath, "r")) == NULL)
		return -1;
	if ((out = fopen(newpath, "w")) == NULL) {
		fclose(in);
		return -1;
	}
	while ((c = getc(in)) != EOF)
		putc(c, out);
	fclose(in);
	return fclose(out);
}

void cd_setdirnames(confdirinfo * cdi)
{
	int i, l;
	strncpy(cdi->home_dir, cdi->env.home, MAXPATHLEN);
	cdi->home_dir[MAXPATHLEN] = '\0';
	cd_purifyname(cdi->home_dir);
	if (cdi->env.config_dir != NULL) {
		if (cdi->env.config_dir[0] == '/') {
			strncpy(cdi->config_dir, cdi->env.config_dir, MAXPATHLEN);
			cdi->config_dir[MAXPATHLEN] = '\0';
			cd_purifyname(cdi->config_dir);
		} else {
			strncpy(cdi->config_dir, cdi->env.home, MAXPATHLEN);
			l = strlen(cdi->config_dir);
			if (l > MAXPATHLEN - 3)
				cd_error("Home directory name too long to create subdirectory names, using HOME as CONFIG_DIR");
			else {
				if (cdi->env.config_dir[0] == '~') {
					i = 1;
					while (i < MAXPATHLEN && cdi->env.config_dir[i] != '\0' && cdi->env.config_dir[i] == '/')
						i++;
				} else
					i = 0;
			}	
			cd_appendsubname(cdi->config_dir, cdi->env.config_dir + i, l);
		}
	}
	if (cdi->env.data_dir != NULL) {
		if (cdi->env.data_dir[0] == '/') {
			strncpy(cdi->data_dir, cdi->env.data_dir, MAXPATHLEN);
			cdi->data_dir[MAXPATHLEN] = '\0';
			cd_purifyname(cdi->data_dir);
		} else {		
			strncpy(cdi->data_dir, cdi->env.home, MAXPATHLEN);
			l = strlen(cdi->data_dir);
			if (l > MAXPATHLEN - 3)
				cd_error("Home directory name too long to create subdirectory names, using HOME as DATA_DIR");
			else {
				if (cdi->env.data_dir[0] == '~') {
					i = 1;
					while (i < MAXPATHLEN && cdi->env.data_dir[i] != '\0' && cdi->env.data_dir[i] == '/')
						i++;
				} else
					i = 0;
			}
			cd_appendsubname(cdi->data_dir, cdi->env.data_dir + i, l);
		}
	}
}

void cd_setcommoninfo(confdirinfo * cdi)
{
	struct stat st;
	if (cdi->info_src == CD_ENVINFO)
		cd_getenvinfo(cdi);
	else if (cdi->info_src == CD_FILEINFO)
		cd_getfileinfo(cdi);
	cd_setdirnames(cdi);
	cdi->config_dir_exists = 0;
	if (cdi->env.config_dir != NULL) {
		if (!stat(cdi->config_dir, &st)) {
			if (!S_ISDIR(st.st_mode))
				fprintf(cd_errchan, "CONFIG_DIR %s is not a directory!\n", cdi->config_dir);
			else
				cdi->config_dir_exists = 1;
		} else
			fprintf(cd_errchan, "CONFIG_DIR %s does not exist!\n", cdi->config_dir);
	}
	cdi->data_dir_exists = 0;
	if (cdi->env.data_dir != NULL) {
		if (!stat(cdi->data_dir, &st)) {
			if (!S_ISDIR(st.st_mode))
				fprintf(cd_errchan, "DATA_DIR %s is not a directory!\n", cdi->data_dir);
			else
				cdi->data_dir_exists = 1;
		} else
			fprintf(cd_errchan, "DATA_DIR %s does not exist!\n", cdi->config_dir);
	}
}

int cd_mkdir(char * pathname, int length_existing, mode_t mode)
{
	int i;
	struct stat st;
	i = length_existing + 1;
	while (pathname[i] != '\0') {
		if (pathname[i] == '/') {
			pathname[i] = '\0';
			if (!stat(pathname, &st)) {
				if (S_ISDIR(st.st_mode)) {
					if ((st.st_mode & 0777) != mode) {
						if (chmod(pathname, mode)) {
							fprintf(cd_errchan, "Could not chmod %s to %o: %s\n", pathname, mode, strerror(errno));
							pathname[i] = '/';
							return -1;
						}
					}
				} else {
					fprintf(cd_errchan, "%s exists, but is not a directory!\n", pathname);
					pathname[i] = '/';
					return -1;
				}
			} else {
				if (mkdir(pathname, mode)) {
					fprintf(cd_errchan, "Could not mkdir %s: %s", pathname, strerror(errno));
					pathname[i] = '/';
					return -1;
				}
			}
			pathname[i] = '/';
		}
		i++;
	}
	if (!stat(pathname, &st)) {
		if (S_ISDIR(st.st_mode)) {
			if ((st.st_mode & 0777) != mode) {
				if (chmod(pathname, mode)) {
					fprintf(cd_errchan, "Could not chmod %s to %o: %s\n", pathname, mode, strerror(errno));
					return -1;
				}
			}
		} else {
			fprintf(cd_errchan, "%s exists, but is not a directory!\n", pathname);
			return -1;
		}
	} else {
		if (mkdir(pathname, mode)) {
			fprintf(cd_errchan, "Could not mkdir %s: %s", pathname, strerror(errno));
			return -1;
		}
	}
	return 0;
}

char * cd_findconfigfile(const char * oldsubdir, const char * oldname, const char * newsubdir, const char * newname, fileinfo * fi)
{
	struct stat st;
	fi->oldpath_moved = 0;
	fi->newpath_exists = 1;
	fi->oldpath_exists = 1;
	if (fi->common_info->config_dir_exists) { /* tylko jeli istnieje $CONFIG_DIR */
		strncpy(fi->newpath, fi->common_info->config_dir, MAXPATHLEN);
		if (newsubdir != NULL) {
			cd_appendsubname(fi->newpath, newsubdir, strlen(fi->newpath));
			cd_purifyname(fi->newpath);
			if (!stat(fi->newpath, &st)) {
				if (!S_ISDIR(st.st_mode)) {
					fprintf(cd_errchan, "%s exists, but is not a directory!\n", fi->newpath);
					fi->newpath_exists = 0;
				}
			} else {
				fprintf(cd_errchan, "Directory %s does not exists, creating it\n", fi->newpath);
				if (cd_mkdir(fi->newpath, strlen(fi->common_info->config_dir), 0700))
					fi->newpath_exists = 0;
			}
		}
		cd_appendsubname(fi->newpath, newname, strlen(fi->newpath));
		cd_purifyname(fi->newpath);
		if (!stat(fi->newpath, &st)) {
			/* nie sprawdzaem istnienia fi->oldpath */
			fi->oldpath_exists = -1;
			return fi->newpath;
		}
	} else
		fi->newpath_exists = 0;
	strncpy(fi->oldpath, fi->common_info->home_dir, MAXPATHLEN);
	if (oldsubdir != NULL) {
		cd_appendsubname(fi->oldpath, oldsubdir, strlen(fi->oldpath));
		cd_purifyname(fi->oldpath);
		if (!stat(fi->oldpath, &st)) {
			if (!S_ISDIR(st.st_mode)) {
				fprintf(cd_errchan, "%s exists, but is not a directory!\n", fi->oldpath);
				fi->oldpath_exists = 0;
			}
		} else {
			if (! fi->common_info->config_dir_exists) {
				fprintf(cd_errchan, "Directory %s does not exist, creating it\n", fi->oldpath);
				if (cd_mkdir(fi->oldpath, strlen(fi->common_info->config_dir), 0700))
					fi->oldpath_exists = 0;
			} else
				fi->oldpath_exists = 0;
		}
	}
	cd_appendsubname(fi->oldpath, oldname, strlen(fi->oldpath));
	cd_purifyname(fi->oldpath);
	if (stat(fi->oldpath, &st))
		fi->oldpath_exists = 0;
	else {
		if (fi->newpath_exists) {
			if (rename(fi->oldpath, fi->newpath)) {
				if (errno != EXDEV) {
					fprintf(cd_errchan, "Could not rename %s to %s, using old location %s\n", fi->oldpath, fi->newpath, fi->oldpath);
					cd_error(strerror(errno));
					fi->newpath_exists = 0;
				} else
					if (cd_rename(fi->oldpath, fi->newpath)) {
						fprintf(cd_errchan, "Could not rename %s to %s, using old location %s\n", fi->oldpath, fi->newpath, fi->oldpath);
						fi->newpath_exists = 0;
					} else /* nie ma ju fi->oldpath */
						fi->oldpath_exists = 0;
			} else {/* nie ma ju fi->oldpath */
				fprintf(cd_errchan, "Moved %s to %s\n", fi->oldpath, fi->newpath);
				fi->oldpath_exists = 0;
			}
		}
	}
	if (fi->newpath_exists) /* jedno z dwojga: nie ma fi->oldpath a istnieje katalog w ktrym mona utworzy goy fi->newpath, albo istnieje sobie ju fi->newpath */
		return fi->newpath;
	/* a tak, to czytaj/utwrz fi->oldpath */
	return fi->oldpath;
}

char * cd_finddatafile(const char * oldsubdir, const char * oldname, const char * newsubdir, const char * newname, fileinfo * fi)
{
	struct stat st;
	fi->oldpath_moved = 0;
	fi->newpath_exists = 1;
	fi->oldpath_exists = 1;
	if (fi->common_info->data_dir_exists) {
		strncpy(fi->newpath, fi->common_info->data_dir, MAXPATHLEN);
		if (newsubdir != NULL) {
			cd_appendsubname(fi->newpath, newsubdir, strlen(fi->newpath));
			cd_purifyname(fi->newpath);
			if (!stat(fi->newpath, &st)) {
				if (!S_ISDIR(st.st_mode)) {
					fprintf(cd_errchan, "%s exists, but is not a directory!\n", fi->newpath);
					fi->newpath_exists = 0;
				}
			} else {
				fprintf(cd_errchan, "Directory %s does not exists, creating it\n", fi->newpath);
				if (cd_mkdir(fi->newpath, strlen(fi->common_info->data_dir), 0700))
					fi->newpath_exists = 0;
			}
		}
		cd_appendsubname(fi->newpath, newname, strlen(fi->newpath));
		cd_purifyname(fi->newpath);
		if (!stat(fi->newpath, &st)) {
			/* nie sprawdzaem istnienia fi->oldpath */
			fi->oldpath_exists = -1;
			return fi->newpath;
		}
	} else
		fi->newpath_exists = 0;
	strncpy(fi->oldpath, fi->common_info->home_dir, MAXPATHLEN);
	if (oldsubdir != NULL) {
		cd_appendsubname(fi->oldpath, oldsubdir, strlen(fi->oldpath));
		cd_purifyname(fi->oldpath);
		if (!stat(fi->oldpath, &st)) {
			if (!S_ISDIR(st.st_mode)) {
				fprintf(cd_errchan, "%s exists, but is not a directory!\n", fi->oldpath);
				fi->oldpath_exists = 0;
			}
		} else {
			if (! fi->common_info->data_dir_exists) {
				fprintf(cd_errchan, "Directory %s does not exist, creating it\n", fi->oldpath);
				if (cd_mkdir(fi->oldpath, strlen(fi->common_info->data_dir), 0700))
					fi->oldpath_exists = 0;
			} else
				fi->oldpath_exists = 0;
		}
	}
	cd_appendsubname(fi->oldpath, oldname, strlen(fi->oldpath));
	cd_purifyname(fi->oldpath);
	if (stat(fi->oldpath, &st))
		fi->oldpath_exists = 0;
	else {
		if (fi->newpath_exists) {
			if (rename(fi->oldpath, fi->newpath)) {
				if (errno != EXDEV) {
					fprintf(cd_errchan, "Could not rename %s to %s: %s", fi->oldpath, fi->newpath, strerror(errno));
					fprintf(cd_errchan, "Using location old %s\n", fi->oldpath);
					fi->newpath_exists = 0;
				} else
					if (cd_rename(fi->oldpath, fi->newpath)) {
						fprintf(cd_errchan, "Could not rename %s to %s, using old location %s\n", fi->oldpath, fi->newpath, fi->oldpath);
						fi->newpath_exists = 0;
					} else /* nie ma ju fi->oldpath */
						fi->oldpath_exists = 0;
			} else {/* nie ma ju fi->oldpath */
				fprintf(cd_errchan, "Moved %s to %s\n", fi->oldpath, fi->newpath);
				fi->oldpath_exists = 0;
			}
		}
	}
	if (fi->newpath_exists) /* jedno z dwojga: nie ma fi->oldpath a istnieje katalog w ktrym mona utworzy goy fi->newpath, albo istnieje sobie ju fi->newpath */
		return fi->newpath;
	/* a tak, to czytaj/utwrz fi->oldpath */
	return fi->oldpath;
}
