#ifndef _CONF_DIR_H
#define _CONF_DIR_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#define CD_FILEINFO 1
#define CD_ENVINFO 2
#define CD_ETCRC ".etcrc"
#define CD_GLOBAL_ETCRC "/etc/etcrc"

typedef struct {
	struct {
		char * config_dir;
		char * data_dir;
		char * home;
	} env;
	char config_dir[MAXPATHLEN + 1];
	char data_dir[MAXPATHLEN + 1];
	char home_dir[MAXPATHLEN + 1];
	int data_dir_exists;
	int config_dir_exists;
	int info_src;
} confdirinfo;

typedef struct {
	confdirinfo * common_info;
	char oldpath[MAXPATHLEN];
	char newpath[MAXPATHLEN];	
	int movable;
	int oldpath_exists;
	int newpath_exists;
	int oldpath_moved;
} fileinfo;

/* ustaw pola rekordu cdi->env na wartoci odpowiednich zmiennych rodowiskowych*/
extern void cd_getenvinfo(confdirinfo * cdi);

/* dzieli lini tekstu na wyrazy, traktujc "aa aa" albo 'aa aa' jako jeden wyraz */
extern int cd_splitline(char * line, char *** tokens, int len);

/* ustaw pola rekordu cdi->env na wartoci odczytane z pliku ~/.etcrc lub /etc/etcrc */
extern int cd_getfileinfo(confdirinfo * cdi);

/* dopisz nazw podkatalogu/pliku do nazwy katalogu o dugoci l */
extern void cd_appendsubname(char topdir[MAXPATHLEN + 1], const char * subname, size_t topdirlen);

/* skopiuj plik ze jednego volumenu na drugi (kiedy rename nie moe) */
extern int cd_rename(const char * oldpath, const char * newpath);

/* stwrz katalog o nazwie pathname, przyjmujc, e pierwsze length_existing znakw w pathname podaje nazw ju istniejcego katalogu, a dalsze podkatalogi maj mie prawa dostpu okrelone przez mode */
extern int cd_mkdir(char * pathname, int length_existing, mode_t mode);

/* ustaw pola rekordu cdi z absolutnymi nazwami odpowiednich katalogw, utworzonych ze zmiennych w cdi->env */
extern void cd_setdirnames(confdirinfo * cdi);

/* usu powtarzajce si znaki '/' w nazwie katalogu/pliku i usu znak '/' z koca nazwy, zwr dugo nowej nazwy (bez '\0') */
extern int cd_purifyname(char dirname[MAXPATHLEN]);

/* nadaj wartoci zmiennym w strukturze cdi */
extern void cd_setcommoninfo(confdirinfo * cdi);

/* zwr najlepsz moliw nazw pliku konfiguracyjnego, sprawd czy istnieje */
extern char * cd_findconfigfile(const char * oldsubdir, const char * oldname, const char * newsubdir, const char * newname, fileinfo * fi);

/* podobnie dla $DATA_DIR */
extern char * cd_finddatafile(const char * oldsubdir, const char * oldname, const char * newsubdir, const char * newname, fileinfo * fi);

#ifdef __cplusplus
} /*  extern "C" */
#endif /* __cplusplus */

#endif /* _CONF_DIR_H */
